/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2016 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.security;

import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.security.MasterEncryptionService;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.layout.EncryptConfigurationInterceptor;
import org.artifactory.logging.sumo.SumoLogicTokenManager;
import org.artifactory.security.crypto.CryptoHelper;
import org.artifactory.security.interceptor.ApiKeysEncryptor;
import org.artifactory.security.interceptor.MissionControlEncryptInterceptor;
import org.artifactory.security.interceptor.StoragePropertiesEncryptInterceptor;
import org.artifactory.security.interceptor.UserPasswordEncryptor;
import org.artifactory.security.log.AuditLogger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

/**
 * Implementation of the master encryption service based on symmetric key and Base58 encoding.
 *
 * @author Yossi Shaul
 */
@Service
public class MasterEncryptionServiceImpl implements MasterEncryptionService {

    @Autowired
    private CentralConfigService centralConfigService;

    @Autowired
    private ApiKeysEncryptor apiKeysEncryptor;

    @Autowired
    private UserPasswordEncryptor userPasswordEncryptor;

    @Autowired
    private SumoLogicTokenManager sumoLogicTokenManager;

    @Autowired
    private AuditLogger auditLog;

    @Override
    public void encrypt() {
        AccessLogger.approved("Encrypting with master encryption key");
        // Create the master key if needed
        if (!CryptoHelper.hasMasterKey()) {
            CryptoHelper.createMasterKeyFile();
        }
        new StoragePropertiesEncryptInterceptor().encryptOrDecryptStoragePropertiesFile(true);
        new MissionControlEncryptInterceptor().encryptOrDecryptMissionControlPropertiesFile(true);
        apiKeysEncryptor.encryptOrDecrypt(true);
        userPasswordEncryptor.encryptOrDecrypt(true);
        sumoLogicTokenManager.encryptOrDecryptAllTokens(true);
        auditLog.configurationEncrypted();
        // config interceptor will encrypt the config before it is saved to the database
        MutableCentralConfigDescriptor mutableDescriptor = centralConfigService.getMutableDescriptor();
        centralConfigService.saveEditedDescriptorAndReload(mutableDescriptor);
    }

    @Override
    public void decrypt() {
        if (!CryptoHelper.hasMasterKey()) {
            throw new IllegalStateException("Cannot decrypt without master key file");
        }
        AccessLogger.approved("Decrypting with master encryption key");
        MutableCentralConfigDescriptor mutableDescriptor = centralConfigService.getMutableDescriptor();
        new StoragePropertiesEncryptInterceptor().encryptOrDecryptStoragePropertiesFile(false);
        new MissionControlEncryptInterceptor().encryptOrDecryptMissionControlPropertiesFile(false);
        apiKeysEncryptor.encryptOrDecrypt(false);
        userPasswordEncryptor.encryptOrDecrypt(false);
        sumoLogicTokenManager.encryptOrDecryptAllTokens(false);
        EncryptConfigurationInterceptor.decrypt(mutableDescriptor);
        CryptoHelper.removeMasterKeyFile();
        auditLog.configurationDecrypted();
        centralConfigService.saveEditedDescriptorAndReload(mutableDescriptor);
    }
}
