/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2016 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.converters;

import org.artifactory.api.common.MoveMultiStatusHolder;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.SecurityService;
import org.artifactory.aql.AqlService;
import org.artifactory.aql.api.domain.sensitive.AqlApiItem;
import org.artifactory.aql.result.AqlEagerResult;
import org.artifactory.aql.result.rows.AqlItem;
import org.artifactory.aql.util.AqlUtils;
import org.artifactory.descriptor.repo.DockerApiVersion;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RepoBaseDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.fs.ItemInfo;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.util.PathUtils;
import org.artifactory.version.CompoundVersionDetails;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContextHolder;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.*;
import static org.artifactory.aql.api.internal.AqlBase.and;

/**
 * @author Nadav Yogev
 */
class DockerManifestImageConverter implements PostInitConverter {
    private static final Logger log = LoggerFactory.getLogger(DockerManifestImageConverter.class);

    @Override
    public void convert(CompoundVersionDetails source, CompoundVersionDetails target) {
        // Get permission to read and move repositories
        ArtifactoryContext artifactoryContext = ContextHelper.get();
        SecurityService securityService = artifactoryContext.beanForType(SecurityService.class);
        Authentication authentication = SecurityContextHolder.getContext().getAuthentication();
        try {
            securityService.authenticateAsSystem();
            convert(artifactoryContext);
        } finally {
            // Restore previous permissions
            SecurityContextHolder.getContext().setAuthentication(authentication);
        }

    }

    private List<String> getLocalDockerRepos(ArtifactoryContext artifactoryContext) {
        Map<String, LocalRepoDescriptor> localRepositoriesMap = artifactoryContext.getCentralConfig().getDescriptor()
                .getLocalRepositoriesMap();
        return new ArrayList<>(localRepositoriesMap.values())
                .stream()
                .filter(r -> RepoType.Docker.equals(r.getType()) && DockerApiVersion.V2.equals(r.getDockerApiVersion()))
                .map(RepoBaseDescriptor::getKey)
                .collect(Collectors.toList());
    }

    /**
     * Find candidates for converter and move isolated images to the correct manifest.json folder
     */
    public void convert(ArtifactoryContext artifactoryContext) {
        List<String> dockers = getLocalDockerRepos(artifactoryContext);
        dockers.forEach((repoKey) -> {
            String path = repoKey + "/*";
            // Find candidates from db fore docker repo
            AqlApiItem query = AqlApiItem.create()
                    .filter(
                            and(
                                    repo().equal(repoKey),
                                    path().matches(path),
                                    type().equal("file"),
                                    depth().greaterEquals(4)
                            )
                    );
            AqlService aqlService = artifactoryContext.beanForType(AqlService.class);
            AqlEagerResult<AqlItem> result = aqlService.executeQueryEager(query);
            // Convert candidates to RepoPath
            Set<RepoPath> repoPaths = result.getResults().stream().map(r -> AqlUtils.fromAql(r).getParent())
                    .collect(Collectors.toSet());
            // Filter docker repos that need to be combined
            repoPaths = repoPaths.stream().filter(r -> isDockerRepoRequireCombine(r, artifactoryContext))
                    .collect(Collectors.toSet());
            // Move images to the correct location (manifest location)
            repoPaths.forEach(r -> move(r, artifactoryContext));
        });
    }

    /**
     * Move files from imageParent to manifest path
     *
     * @param imageParent parent path to move from
     */
    private void move(RepoPath imageParent, ArtifactoryContext artifactoryContext) {
        RepoPath manifestParent = getManifestRepoPath(imageParent);
        RepositoryService repositoryService = artifactoryContext.getRepositoryService();
        List<ItemInfo> children = repositoryService.getChildren(imageParent);
        log.debug("Merging {} files from {} to {}", children.size(), imageParent, manifestParent);
        children.forEach(child -> {
            MoveMultiStatusHolder moveMultiStatusHolder = repositoryService
                    .move(child.getRepoPath(), manifestParent, false, true, true);
            if (moveMultiStatusHolder.isError()) {
                log.error("Failed to move {} to {}. Reason: {}", child, manifestParent,
                        moveMultiStatusHolder.getLastError().getMessage());
            }
        });
        while (repositoryService.getArtifactCount(imageParent) == 0) {
            // remove folder
            log.debug("Removing empty folder {}", imageParent);
            repositoryService.undeploy(imageParent);
            if (imageParent != null) {
                imageParent = imageParent.getParent();
            }
        }
    }

    /**
     * Checks if images and manifest.json are separated and needs merging
     *
     * @param repoPath path of image files folder
     * @return returns true if images and manifest are separated and need combine
     */
    private boolean isDockerRepoRequireCombine(RepoPath repoPath, ArtifactoryContext artifactoryContext) {
        RepoPath manifestRepoPath = getManifestRepoPath(repoPath);

        boolean isManifestExists = isManifestExists(manifestRepoPath, artifactoryContext);
        boolean isImagesIsolated = isImagesIsolated(repoPath, artifactoryContext);
        if (isImagesIsolated && isManifestExists) {
            log.info("Merging manifest.json in {} with its corresponding layers.", manifestRepoPath);
            return true;
        }
        return false;
    }

    /**
     * Build manifest probable path, without the starting path element
     *
     * @param repoPath path of possible separated images
     * @return manifest.json probable path
     */
    private RepoPath getManifestRepoPath(RepoPath repoPath) {
        String root = PathUtils.getFirstPathElement(repoPath.getPath());
        String manifestPath = repoPath.getPath().replaceFirst(root, "");
        manifestPath = PathUtils.removeDuplicateSlashes(manifestPath) + "/";
        return RepoPathFactory.create(repoPath.getRepoKey(), manifestPath);
    }

    /**
     * Checks if there is no manifest.json in images folder
     *
     * @param manifestRepoPath RepoPath of images
     * @return true if no manifest.json file exists in folder
     */
    private boolean isImagesIsolated(RepoPath manifestRepoPath, ArtifactoryContext artifactoryContext) {
        List<ItemInfo> children = artifactoryContext.getRepositoryService().getChildren(manifestRepoPath);
        return children.stream().filter(r -> r.getName().equals("manifest.json")).count() == 0;
    }

    /**
     * Checks if repoPath contains manifest.json
     *
     * @param repoPath RepoPath of probable manifest.json
     * @return true if manifest.json exists in path
     */
    private boolean isManifestExists(RepoPath repoPath, ArtifactoryContext artifactoryContext) {
        List<ItemInfo> children = artifactoryContext.getRepositoryService().getChildren(repoPath);
        return children.stream().filter(r -> r.getName().equals("manifest.json")).count() == 1;
    }
}
