/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2016 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.converters;

import org.artifactory.version.ArtifactoryVersion;
import org.artifactory.version.CompoundVersionDetails;
import org.artifactory.version.VersionComparator;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * @author Nadav Yogev
 */
public enum PostInitVersions {
    v100(ArtifactoryVersion.v4110, ArtifactoryVersion.v4121, new DockerManifestImageConverter());

    private final VersionComparator comparator;
    private final PostInitConverter[] converters;

    /**
     * @param from       The artifactory version this config version was first used
     * @param until      The artifactory version this config was last used in (inclusive)
     * @param converters A list of converters to use to move from <b>this</b> config version to the <b>next</b> config
     *                   version
     */
    PostInitVersions(ArtifactoryVersion from, ArtifactoryVersion until,
                     PostInitConverter... converters) {
        this.comparator = new VersionComparator(from, until);
        this.converters = converters;
    }

    public static PostInitVersions getCurrent() {
        PostInitVersions[] versions = PostInitVersions.values();
        return versions[versions.length - 1];
    }

    public boolean isCurrent() {
        return comparator.isCurrent();
    }

    public void convert(CompoundVersionDetails from, CompoundVersionDetails until) {
        // First create the list of converters to apply
        List<PostInitConverter> converters = new ArrayList<>();
        // All converters of versions between desired range needs to be executed in sequence
        PostInitVersions[] versions = PostInitVersions.values();
        for (PostInitVersions version : versions) {
            if (from.getVersion().after(version.comparator.getFrom()) && from.getVersion().before(version.comparator.getUntil())) {
                converters.addAll(Arrays.asList(version.getConverters()));
            }
        }
        // Apply converters
        converters.forEach(converter -> converter.convert(from, until));
    }

    public PostInitConverter[] getConverters() {
        return converters;
    }
}
