/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2016 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.security.crypto;

import org.artifactory.common.ArtifactoryHome;
import org.artifactory.common.ConstantValues;
import org.jfrog.security.crypto.DummyEncryptionWrapper;
import org.jfrog.security.crypto.EncodingType;
import org.jfrog.security.crypto.EncryptionWrapper;
import org.jfrog.security.crypto.EncryptionWrapperFactory;
import org.jfrog.security.file.SecurityFolderHelper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;


/**
 * Helper class for encrypting/decrypting passwords.
 *
 * @author Yossi Shaul
 */
public abstract class CryptoHelper {
    private static final Logger log = LoggerFactory.getLogger(CryptoHelper.class);

    private static final EncryptionWrapper DUMMY_WRAPPER = new DummyEncryptionWrapper();

    private CryptoHelper() {
        // utility class
    }

    public static EncryptionWrapper getMasterKeyWrapper() {
        ArtifactoryHome home = ArtifactoryHome.get();
        EncryptionWrapper wrapper = home.getMasterEncryptionWrapper();
        if (wrapper == null) {
            File masterKeyFile = getMasterKeyFile();
            if (masterKeyFile.exists()) {
                wrapper = EncryptionWrapperFactory.createMasterWrapper(masterKeyFile);
                home.setMasterEncryptionWrapper(wrapper);
            } else {
                wrapper = DUMMY_WRAPPER;
            }
        }
        return wrapper;
    }

    public static boolean isMasterEncrypted(String in) {
        return EncodingType.ARTIFACTORY_MASTER.isEncodedByMe(in);
    }

    public static boolean isEncryptedUserPassword(String in) {
        return EncodingType.ARTIFACTORY_PASSWORD.isEncodedByMe(in);
    }

    public static boolean isApiKey(String in) {
        return EncodingType.ARTIFACTORY_API_KEY.isEncodedByMe(in);
    }

    public static String encryptIfNeeded(String password) {
        // If password null => no encryption
        if (password == null || password.length() == 0) {
            return password;
        }
        // If no master key encryption => no encryption
        File keyFile = getMasterKeyFile();
        if (!keyFile.exists()) {
            return password;
        }

        // If already encrypted => no encrypt again
        if (isMasterEncrypted(password)) {
            return password;
        }

        return getMasterKeyWrapper().encryptIfNeeded(password);
    }

    public static String decryptIfNeeded(String password) {
        if (isMasterEncrypted(password)) {
            File keyFile = getMasterKeyFile();
            if (!keyFile.exists()) {
                throw new IllegalArgumentException("The Password is encrypted.\n" +
                        "And no Master Key file found at " + keyFile.getAbsolutePath());
            }
            return getMasterKeyWrapper().decryptIfNeeded(password);
        }
        return password;
    }

    /**
     * Renames the master key file, effectively disabling encryption.
     */
    public static void removeMasterKeyFile() {
        File keyFile = getMasterKeyFile();
        SecurityFolderHelper.removeKeyFile(keyFile);
        unsetMasterEncryptionWrapper();
    }

    public static void unsetMasterEncryptionWrapper() {
        ArtifactoryHome.get().setMasterEncryptionWrapper(null);
    }

    /**
     * Creates a master encryption key file. Throws an exception if the key file already exists of on any failure with
     * file or key creation.
     */
    public static void createMasterKeyFile() {
        File keyFile = getMasterKeyFile();
        SecurityFolderHelper.createKeyFile(keyFile);
        ArtifactoryHome.get().setMasterEncryptionWrapper(EncryptionWrapperFactory.createMasterWrapper(keyFile));
    }

    /**
     * @return Master encryption file configured location. The file might not exist.
     */
    public static File getMasterKeyFile() {
        ArtifactoryHome home = ArtifactoryHome.get();
        String keyFileLocation = ConstantValues.securityMasterKeyLocation.getString();
        File keyFile = new File(keyFileLocation);
        if (!keyFile.isAbsolute()) {
            keyFile = new File(home.getHaAwareEtcDir(), keyFileLocation);
        }
        return keyFile;
    }

    public static boolean hasMasterKey() {
        return getMasterKeyFile().exists();
    }
}
