/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2016 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.util.bearer;

import org.apache.http.Header;
import org.apache.http.HttpRequest;
import org.apache.http.auth.AUTH;
import org.apache.http.auth.AuthenticationException;
import org.apache.http.auth.Credentials;
import org.apache.http.impl.auth.RFC2617Scheme;
import org.apache.http.message.BufferedHeader;
import org.apache.http.protocol.BasicHttpContext;
import org.apache.http.protocol.HttpContext;
import org.apache.http.util.CharArrayBuffer;
import org.artifactory.addon.docker.rest.DockerRemoteTokenProvider;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.repo.distribution.auth.BintrayTokenProvider;

import java.util.Optional;

/**
 * Bearer authentication scheme as defined in RFC 2617
 *
 * @author Shay Yaakov
 */
public class BearerScheme extends RFC2617Scheme {
    private String repoKey;
    private TokenProvider tokenProvider;

    public BearerScheme(String repoKey) {
        this.repoKey = repoKey;
        tokenProvider = getTokenProviderByRepoType(repoKey);
    }

    @Override
    public String getSchemeName() {
        return "bearer";
    }

    @Override
    public boolean isConnectionBased() {
        return false;
    }

    @Override
    public boolean isComplete() {
        return true;
    }

    @Override
    public Header authenticate(Credentials dummyCredentials, HttpRequest request) throws AuthenticationException {
        return authenticate(dummyCredentials, request, new BasicHttpContext());
    }

    @Override
    public Header authenticate(Credentials dummyCredentials, HttpRequest request, HttpContext context)
            throws AuthenticationException {
        String token = tokenProvider.getToken(getParameters(),
                request.getRequestLine().getMethod(),
                request.getRequestLine().getUri(),
                repoKey);
        final CharArrayBuffer buffer = new CharArrayBuffer(32);
        buffer.append(AUTH.WWW_AUTH_RESP);
        buffer.append(": Bearer ");
        buffer.append(token);
        return new BufferedHeader(buffer);
    }

    private TokenProvider getTokenProviderByRepoType(String repoKey) {
        TokenProvider provider;
        RepoType type = Optional.ofNullable(ContextHelper.get().beanForType(RepositoryService.class)
                .repoDescriptorByKey(repoKey))
                .orElseThrow(() -> new RuntimeException("No such repository " + repoKey))
                .getType();
        switch (type) {
            case Distribution:
                provider = ContextHelper.get().beanForType(BintrayTokenProvider.class);
                break;
            case Docker:
                provider = ContextHelper.get().beanForType(DockerRemoteTokenProvider.class);
                break;
            default:
                throw new IllegalArgumentException("Token Authentication is not available for repositories of type "
                        + type);
        }
        return provider;
    }
}
