/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2016 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.storage.db.version;

import com.google.common.collect.Lists;
import org.artifactory.storage.db.DbType;
import org.artifactory.storage.db.util.JdbcHelper;
import org.artifactory.storage.db.version.converter.DBConverter;
import org.artifactory.storage.db.version.converter.DBSqlConverter;
import org.artifactory.version.ArtifactoryVersion;
import org.artifactory.version.VersionComparator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;

/**
 * Artifactory DB version
 */
public enum ArtifactoryDBVersion {
    v100(ArtifactoryVersion.v300, ArtifactoryVersion.v304),
    v101(ArtifactoryVersion.v310, ArtifactoryVersion.v310, new DBSqlConverter("v310")),
    v102(ArtifactoryVersion.v311, ArtifactoryVersion.v402, new DBSqlConverter("v311")),
    v103(ArtifactoryVersion.v410, ArtifactoryVersion.v412, new DBSqlConverter("v410")),
    v104(ArtifactoryVersion.v420, ArtifactoryVersion.v431, new DBSqlConverter("v420")),
    v106(ArtifactoryVersion.v432, ArtifactoryVersion.v433),
    v107(ArtifactoryVersion.v440, ArtifactoryVersion.v440, new DBSqlConverter("v440")),
    v108(ArtifactoryVersion.v441, ArtifactoryVersion.getCurrent(), new DBSqlConverter("v441"));

    private static final Logger log = LoggerFactory.getLogger(ArtifactoryDBVersion.class);


    private final VersionComparator comparator;
    private final DBConverter[] converters;

    ArtifactoryDBVersion(ArtifactoryVersion from, ArtifactoryVersion until, DBConverter... converters) {
        this.comparator = new VersionComparator(from, until);
        this.converters = converters;
    }

    public static ArtifactoryDBVersion getLast() {
        ArtifactoryDBVersion[] versions = ArtifactoryDBVersion.values();
        return versions[versions.length - 1];
    }

    public static void convert(ArtifactoryVersion from, JdbcHelper jdbcHelper, DbType dbType) {
        // All converters of versions above me needs to be executed in sequence
        List<DBConverter> converters = Lists.newArrayList();
        for (ArtifactoryDBVersion version : ArtifactoryDBVersion.values()) {
            if (version.comparator.isAfter(from) && !version.comparator.supports(from)) {
                for (DBConverter dbConverter : version.getConverters()) {
                    converters.add(dbConverter);
                }
            }
        }

        if (converters.isEmpty()) {
            log.debug("No database converters found between version {} and {}", from, ArtifactoryVersion.getCurrent());
        } else {
            log.info("Starting database conversion from {} to {}", from, ArtifactoryVersion.getCurrent());
            for (DBConverter converter : converters) {
                converter.convert(jdbcHelper, dbType);
            }
            log.info("Finished database conversion from {} to {}", from, ArtifactoryVersion.getCurrent());
        }
    }

    public DBConverter[] getConverters() {
        return converters;
    }

    public VersionComparator getComparator() {
        return comparator;
    }
}
