/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2016 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.actions;

import com.google.common.collect.Lists;
import org.artifactory.addon.AddonType;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.bower.BowerAddon;
import org.artifactory.addon.composer.ComposerAddon;
import org.artifactory.addon.pypi.PypiAddon;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.ivy.IvyNaming;
import org.artifactory.mime.MavenNaming;
import org.artifactory.mime.NamingUtils;
import org.artifactory.repo.RepoPath;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;

/**
 * Builds tabs list to be added to a tree node
 *
 * @author Shay Yaakov
 */
@Component
public class TreeNodeTabsPopulator {

    @Autowired
    private RepositoryService repoService;

    @Autowired
    private AuthorizationService authService;

    @Autowired
    private AddonsManager addonsManager;

    public void populateForRepository(TabsAndActions tabsAndActions) {
        List<TabOrAction> tabs = Lists.newArrayList();
        RepoPath repoPath = tabsAndActions.getRepoPath();
        String repoType = tabsAndActions.getRepoType();
        if ("remote".equals(repoType) || "virtual".equals(repoType) || "trash".equals(repoType)) {
            tabs.add(general());
        } else {
            tabs.add(general());
            tabs.add(effectivePermissions(repoPath));
            tabs.add(properties());
            tabs.add(watch(repoPath));
        }
        tabsAndActions.setTabs(tabs.stream().filter(Objects::nonNull).collect(Collectors.toList()));
    }

    public void populateForFolder(TabsAndActions tabsAndActions) {
        List<TabOrAction> tabs = Lists.newArrayList();
        RepoPath repoPath = tabsAndActions.getRepoPath();
        String repoType = tabsAndActions.getRepoType();
        if ("remote".equals(repoType) || "virtual".equals(repoType) || "trash".equals(repoType)) {
            tabs.add(general());
        } else if ("distribution".equals(repoType)) {
            tabs.add(general());
            tabs.add(effectivePermissions(repoPath));
            tabs.add(properties());
            tabs.add(watch(repoPath));
        } else {
            tabs.add(general());
            tabs.addAll(dockerV1(repoPath));
            tabs.add(dockerV2(repoPath));
            tabs.add(effectivePermissions(repoPath));
            tabs.add(properties());
            tabs.add(watch(repoPath));
        }
        tabsAndActions.setTabs(tabs.stream().filter(Objects::nonNull).collect(Collectors.toList()));
    }

    public void populateForFile(TabsAndActions tabsAndActions) {
        List<TabOrAction> tabs = Lists.newArrayList();
        RepoPath repoPath = tabsAndActions.getRepoPath();
        String repoType = tabsAndActions.getRepoType();
        if ("remote".equals(repoType) || "virtual".equals(repoType) || "trash".equals(repoType)) {
            tabs.add(general());
        } else if ("distribution".equals(repoType)) {
            tabs.add(general());
            tabs.add(effectivePermissions(repoPath));
            tabs.add(properties());
            tabs.add(watch(repoPath));
        } else {
            tabs.add(general());
            tabs.add(pom(repoPath));
            tabs.add(xml(repoPath));
            tabs.add(rpm(repoPath));
            tabs.add(nuget(repoPath));
            tabs.add(gem(repoPath));
            tabs.add(npm(repoPath));
            tabs.add(bower(repoPath));
            tabs.add(composer(repoPath));
            tabs.add(pypi(repoPath));
            tabs.add(effectivePermissions(repoPath));
            tabs.add(properties());
            tabs.add(watch(repoPath));
            tabs.add(builds());
            tabs.add(blackduck());
        }
        tabsAndActions.setTabs(tabs.stream().filter(Objects::nonNull).collect(Collectors.toList()));
    }

    private TabOrAction builds() {
        if (!authService.isAnonUserAndAnonBuildInfoAccessDisabled()) {
            return new TabOrAction("Builds");
        }
        return null;
    }

    private TabOrAction blackduck() {
        return new TabOrAction("BlackDuck");
    }

    private TabOrAction pom(RepoPath repoPath) {
        if (MavenNaming.isPom((repoPath.getName()))) {
            return new TabOrAction("PomView");
        }
        return null;
    }

    private TabOrAction xml(RepoPath repoPath) {
        if (NamingUtils.isXml(repoPath.getName()) && !MavenNaming.isPom((repoPath.getName()))) {
            return IvyNaming.isIvyFileName(repoPath.getName()) ? new TabOrAction("IVYXml") : new TabOrAction("GeneralXml");
        }
        return null;
    }

    private TabOrAction rpm(RepoPath repoPath) {
        if (repoPath.getName().endsWith(".rpm") && isRepoSupportType(RepoType.YUM, repoPath)
                && addonsManager.isAddonSupported(AddonType.YUM)) {
            return new TabOrAction("Rpm");
        }
        return null;
    }

    private TabOrAction nuget(RepoPath repoPath) {
        if (repoPath.getName().endsWith(".nupkg") && isRepoSupportType(RepoType.NuGet, repoPath)
                && addonsManager.isAddonSupported(AddonType.NUGET)) {
            return new TabOrAction("NuPkgInfo");
        }
        return null;
    }

    private TabOrAction gem(RepoPath repoPath) {
        if (repoPath.getName().endsWith(".gem") && isRepoSupportType(RepoType.Gems, repoPath)
                && addonsManager.isAddonSupported(AddonType.GEMS)) {
            return new TabOrAction("RubyGems");
        }
        return null;
    }

    private TabOrAction npm(RepoPath repoPath) {
        if (repoPath.getName().endsWith(".tgz") && isRepoSupportType(RepoType.Npm, repoPath)
                && addonsManager.isAddonSupported(AddonType.NPM)) {
            return new TabOrAction("NpmInfo");
        }
        return null;
    }

    private TabOrAction bower(RepoPath repoPath) {
        boolean isBowerFile = addonsManager.addonByType(BowerAddon.class).isBowerFile(repoPath.getName());
        if (isBowerFile && isRepoSupportType(RepoType.Bower, repoPath) && addonsManager.isAddonSupported(AddonType.BOWER)) {
            return new TabOrAction("BowerInfo");
        }
        return null;
    }

    private TabOrAction composer(RepoPath repoPath) {
        boolean isComposerFile = addonsManager.addonByType(ComposerAddon.class).isComposerSupportedExtension(repoPath.getName());
        if (isComposerFile && isRepoSupportType(RepoType.Composer, repoPath) &&
                addonsManager.isAddonSupported(AddonType.COMPOSER)) {
            return new TabOrAction("ComposerInfo");
        }
        return null;
    }

    private TabOrAction pypi(RepoPath repoPath) {
        boolean isPypiFile = addonsManager.addonByType(PypiAddon.class).isPypiFile(repoPath);
        if (isPypiFile && isRepoSupportType(RepoType.Pypi, repoPath) && addonsManager.isAddonSupported(AddonType.PYPI)) {
            return new TabOrAction("PyPIInfo");
        }
        return null;
    }

    private List<TabOrAction> dockerV1(RepoPath repoPath) {
        if (isDockerFileTypeAndSupported(repoPath)) {
            return Lists.newArrayList(new TabOrAction("DockerInfo"), new TabOrAction("DockerAncestryInfo"));
        }
        return Lists.newArrayList();
    }

    private TabOrAction dockerV2(RepoPath repoPath) {
        if (isDockerManifestFolder(repoPath)) {
            return new TabOrAction("DockerV2Info");
        }
        return null;
    }

    private boolean isDockerFileTypeAndSupported(RepoPath repoPath) {
        boolean isDockerEnabled = isRepoSupportType(RepoType.Docker, repoPath) && addonsManager.isAddonSupported(AddonType.DOCKER);
        return isDockerEnabled && repoService.getChildrenNames(repoPath).contains("json.json");
    }

    private boolean isDockerManifestFolder(RepoPath repoPath) {
        boolean isDockerEnabled = isRepoSupportType(RepoType.Docker, repoPath) && addonsManager.isAddonSupported(AddonType.DOCKER);
        return isDockerEnabled && repoService.getChildrenNames(repoPath).contains("manifest.json");
    }

    boolean isRepoSupportType(RepoType type, RepoPath repoPath) {
        LocalRepoDescriptor descriptor = repoService.localCachedOrDistributionRepoDescriptorByKey(repoPath.getRepoKey());
        return isRepoSupportType(type, descriptor);
    }

    private boolean isRepoSupportType(RepoType type, LocalRepoDescriptor descriptor) {
        return descriptor != null && (descriptor.getType().equals(type) || isDistributionRepo(descriptor));
    }

    private boolean isDistributionRepo(LocalRepoDescriptor descriptor) {
        return descriptor.getType().equals(RepoType.Distribution);
    }

    private TabOrAction general() {
        return new TabOrAction("General");
    }

    private TabOrAction effectivePermissions(RepoPath repoPath) {
        if (authService.canManage(repoPath)) {
            return new TabOrAction("EffectivePermission");
        }
        return null;
    }

    private TabOrAction properties() {
        return new TabOrAction("Properties");
    }

    private TabOrAction watch(RepoPath repoPath) {
        if (authService.canManage(repoPath)) {
            return new TabOrAction("Watch");
        }
        return null;
    }
}
