/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2016 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.service.admin.configuration.repositories;

import org.apache.http.HttpStatus;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.config.XrayOperationException;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.common.service.RestService;
import org.artifactory.ui.rest.model.admin.configuration.repository.RepositoryConfigModel;
import org.artifactory.ui.rest.service.admin.configuration.repositories.util.UpdateRepoConfigHelper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import static org.artifactory.util.ExceptionUtils.getCauseOfTypes;

/**
 * @author Dan Feldman
 * @author Aviad Shikloshi
 */
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
public class UpdateRepositoryConfigService implements RestService<RepositoryConfigModel> {
    private static final Logger log = LoggerFactory.getLogger(UpdateRepositoryConfigService.class);

    @Autowired
    private CentralConfigService configService;

    @Autowired
    private UpdateRepoConfigHelper updater;

    @Override
    public void execute(ArtifactoryRestRequest request, RestResponse response) {
        RepositoryConfigModel model = (RepositoryConfigModel) request.getImodel();
        MutableCentralConfigDescriptor configDescriptor = configService.getMutableDescriptor();
        log.debug("Creating descriptor from received model");
        String repoKey = model.getGeneral().getRepoKey();
        if (!configDescriptor.isRepositoryExists(repoKey)) {
            response.error("Repository '" + repoKey + "' doesn't exist.").responseCode(HttpStatus.SC_BAD_REQUEST);
            return;
        }
        log.info("Updating repository {}", repoKey);
        try {
            model.updateRepo(updater);
            response.info("Successfully updated repository '" + repoKey + "'");
        } catch (Exception e) {
            log.debug("Failed to update repository: ", e);

            // Specific error for xray
            Throwable xrayUnsupportedException = getCauseOfTypes(e, XrayOperationException.class);
            if (xrayUnsupportedException != null) {
                String message =
                        "Unable to save Xray config. This may be related to the current available version which" +
                                " does not support the block downloads feature or to other connectivity issues";
                log.error(message + ". " + xrayUnsupportedException.getMessage());
                response.error(message + ". " + "Check the log for more details.");
            } else {
                response.error("Failed to update repository '" + repoKey + "': " + e.getMessage());
            }
        }
    }
}
